# Copyright (c) OpenMMLab. All rights reserved.
from . import base_dense_head  # noqa: F401,F403
from . import centernet_head  # noqa: F401,F403
from . import detr_head  # noqa: F401,F403
from . import fovea_head  # noqa: F401,F403
from . import gfl_head  # noqa: F401,F403
from . import reppoints_head  # noqa: F401,F403
from . import rpn_head  # noqa: F401,F403
from . import rtmdet_head  # noqa: F401,F403
from . import rtmdet_ins_head  # noqa: F401,F403
from . import solo_head  # noqa: F401,F403
from . import solov2_head  # noqa: F401,F403
from . import yolo_head  # noqa: F401,F403
from . import yolox_head  # noqa: F401,F403
