from .bot_sort import BOTSORT
from .byte_tracker import BYTETracker
from .track import Tracker, TRACKER_MAP


# from .track import register_tracker

__all__ = "Tracker", "TRACKER_MAP", "BOTSORT", "BYTETracker"  # allow simpler import
